%% Loading the data
clear
clc
load('wafer_maps.mat')

n_center = size(center, 4); n_center
n_donut = size(donut, 4); n_donut
n_edge_loc = size(edge_loc, 4); n_edge_loc
imshow(center(:, :, :, 2))

%% Creating the label vectors
y_center = categorical(zeros(size(center, 4), 1));
y_donut = categorical(ones(size(donut, 4), 1));
y_edge_loc = categorical(ones(size(edge_loc, 4), 1) * 2);

%% Preparing training, validation, and test sets

rng(3)
ratio_train = 0.5;

idx = randperm(n_center);
idx_train = idx(1:floor(ratio_train * n_center));
idx_valid = ...
    idx(floor(ratio_train * n_center) + 1:floor((ratio_train + ...
    (1 - ratio_train) / 2) * n_center));
idx_test = ...
    idx(floor((ratio_train + (1 - ratio_train) / 2) * n_center) + 1:end);
x_center_train = center(:, :, :, idx_train);
x_center_valid = center(:, :, :, idx_valid);
x_center_test = center(:, :, :, idx_test);
y_center_train = y_center(idx_train);
y_center_valid = y_center(idx_valid);
y_center_test = y_center(idx_test);

idx = randperm(n_donut);
idx_train = idx(1:floor(ratio_train * n_donut));
idx_valid = ...
    idx(floor(ratio_train * n_donut) + 1:floor((ratio_train + ...
    (1 - ratio_train) / 2) * n_donut));
idx_test = ...
    idx(floor((ratio_train + (1 - ratio_train) / 2) * n_donut) + 1:end);
x_donut_train = donut(:, :, :, idx_train);
x_donut_valid = donut(:, :, :, idx_valid);
x_donut_test = donut(:, :, :, idx_test);
y_donut_train = y_donut(idx_train);
y_donut_valid = y_donut(idx_valid);
y_donut_test = y_donut(idx_test);

idx = randperm(n_edge_loc);
idx_train = idx(1:floor(ratio_train * n_edge_loc));
idx_valid = ...
    idx(floor(ratio_train * n_edge_loc) + 1:floor((ratio_train + ...
    (1 - ratio_train) / 2) * n_edge_loc));
idx_test = ...
    idx(floor((ratio_train + (1 - ratio_train) / 2) * n_edge_loc) + 1:end);
x_edge_loc_train = edge_loc(:, :, :, idx_train);
x_edge_loc_valid = edge_loc(:, :, :, idx_valid);
x_edge_loc_test = edge_loc(:, :, :, idx_test);
y_edge_loc_train = y_edge_loc(idx_train);
y_edge_loc_valid = y_edge_loc(idx_valid);
y_edge_loc_test = y_edge_loc(idx_test);

x_train = cat(4, x_center_train, x_donut_train, x_edge_loc_train);
x_valid = cat(4, x_center_valid, x_donut_valid, x_edge_loc_valid);
x_test = cat(4, x_center_test, x_donut_test, x_edge_loc_test);
y_train = cat(1, y_center_train, y_donut_train, y_edge_loc_train);
y_valid = cat(1, y_center_valid, y_donut_valid, y_edge_loc_valid);
y_test = cat(1, y_center_test, y_donut_test, y_edge_loc_test);
length(y_test)
%% Model definition and training

rng(3)
layers = [
    imageInputLayer([32 32 1])
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,64,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
    	
    fullyConnectedLayer(3)
    softmaxLayer
    classificationLayer];

miniBatchSize = 64;
options = trainingOptions( 'adam', 'MiniBatchSize', miniBatchSize, ...
    'MaxEpochs', 10, 'Shuffle', 'every-epoch', ...
    'ValidationData', {x_valid, y_valid}, 'ValidationFrequency', 30, ...
    'Plots', 'training-progress',...
    'InitialLearnRate',0.005);

net = trainNetwork(x_train, y_train, layers, options);

%% Testing the model

y_pred = classify(net, x_test);
test_accuracy = sum(y_pred == y_test) / numel(y_test);
confusionchart(y_test, y_pred)
test_accuracy
%% Accuracy vs. percentage data for training

% Remove plot from training options
options = trainingOptions( 'adam', 'MiniBatchSize', miniBatchSize, ...
    'MaxEpochs', 10, 'Shuffle', 'every-epoch', ...
    'ValidationData', {x_valid, y_valid}, 'ValidationFrequency', 30, 'InitialLearnRate',0.005);

acc = [];
for ratio_train = 0.05:0.05:0.8
    idx = randperm(n_center);
    idx_train = idx(1:floor(ratio_train * n_center));
    idx_valid = ...
        idx(floor(ratio_train * n_center) + 1:floor((ratio_train + ...
        (1 - ratio_train) / 2) * n_center));
    idx_test = ...
        idx(floor((ratio_train + (1 - ratio_train) / 2) * n_center) + 1:end);
    x_center_train = center(:, :, :, idx_train);
    x_center_valid = center(:, :, :, idx_valid);
    x_center_test = center(:, :, :, idx_test);
    y_center_train = y_center(idx_train);
    y_center_valid = y_center(idx_valid);
    y_center_test = y_center(idx_test);

    idx = randperm(n_donut);
    idx_train = idx(1:floor(ratio_train * n_donut));
    idx_valid = ...
        idx(floor(ratio_train * n_donut) + 1:floor((ratio_train + ...
        (1 - ratio_train) / 2) * n_donut));
    idx_test = ...
        idx(floor((ratio_train + (1 - ratio_train) / 2) * n_donut) + 1:end);
    x_donut_train = donut(:, :, :, idx_train);
    x_donut_valid = donut(:, :, :, idx_valid);
    x_donut_test = donut(:, :, :, idx_test);
    y_donut_train = y_donut(idx_train);
    y_donut_valid = y_donut(idx_valid);
    y_donut_test = y_donut(idx_test);

    idx = randperm(n_edge_loc);
    idx_train = idx(1:floor(ratio_train * n_edge_loc));
    idx_valid = ...
        idx(floor(ratio_train * n_edge_loc) + 1:floor((ratio_train + ...
        (1 - ratio_train) / 2) * n_edge_loc));
    idx_test = ...
        idx(floor((ratio_train + (1 - ratio_train) / 2) * n_edge_loc) + 1:end);
    x_edge_loc_train = edge_loc(:, :, :, idx_train);
    x_edge_loc_valid = edge_loc(:, :, :, idx_valid);
    x_edge_loc_test = edge_loc(:, :, :, idx_test);
    y_edge_loc_train = y_edge_loc(idx_train);
    y_edge_loc_valid = y_edge_loc(idx_valid);
    y_edge_loc_test = y_edge_loc(idx_test);

    x_train = cat(4, x_center_train, x_donut_train, x_edge_loc_train);
    x_valid = cat(4, x_center_valid, x_donut_valid, x_edge_loc_valid);
    x_test = cat(4, x_center_test, x_donut_test, x_edge_loc_test);
    y_train = cat(1, y_center_train, y_donut_train, y_edge_loc_train);
    y_valid = cat(1, y_center_valid, y_donut_valid, y_edge_loc_valid);
    y_test = cat(1, y_center_test, y_donut_test, y_edge_loc_test);
    
    net = trainNetwork(x_train, y_train, layers, options);
    y_pred = classify(net, x_test);
    test_accuracy = sum(y_pred == y_test) / numel(y_test);
    acc(end + 1) = test_accuracy;
end

scatter(0.05:0.05:0.8, acc)

%% Accuracy vs. percentage data for training with data augmentation

acc_aug = [];
for ratio_train = 0.05:0.05:0.8
    idx = randperm(n_center);
    idx_train = idx(1:floor(ratio_train * n_center));
    idx_valid = ...
        idx(floor(ratio_train * n_center) + 1:floor((ratio_train + ...
        (1 - ratio_train) / 2) * n_center));
    idx_test = ...
        idx(floor((ratio_train + (1 - ratio_train) / 2) * n_center) + 1:end);
    x_center_train = center(:, :, :, idx_train);
    x_center_valid = center(:, :, :, idx_valid);
    x_center_test = center(:, :, :, idx_test);
    y_center_train = y_center(idx_train);
    y_center_valid = y_center(idx_valid);
    y_center_test = y_center(idx_test);

    idx = randperm(n_donut);
    idx_train = idx(1:floor(ratio_train * n_donut));
    idx_valid = ...
        idx(floor(ratio_train * n_donut) + 1:floor((ratio_train + ...
        (1 - ratio_train) / 2) * n_donut));
    idx_test = ...
        idx(floor((ratio_train + (1 - ratio_train) / 2) * n_donut) + 1:end);
    x_donut_train = donut(:, :, :, idx_train);
    x_donut_valid = donut(:, :, :, idx_valid);
    x_donut_test = donut(:, :, :, idx_test);
    y_donut_train = y_donut(idx_train);
    y_donut_valid = y_donut(idx_valid);
    y_donut_test = y_donut(idx_test);

    idx = randperm(n_edge_loc);
    idx_train = idx(1:floor(ratio_train * n_edge_loc));
    idx_valid = ...
        idx(floor(ratio_train * n_edge_loc) + 1:floor((ratio_train + ...
        (1 - ratio_train) / 2) * n_edge_loc));
    idx_test = ...
        idx(floor((ratio_train + (1 - ratio_train) / 2) * n_edge_loc) + 1:end);
    x_edge_loc_train = edge_loc(:, :, :, idx_train);
    x_edge_loc_valid = edge_loc(:, :, :, idx_valid);
    x_edge_loc_test = edge_loc(:, :, :, idx_test);
    y_edge_loc_train = y_edge_loc(idx_train);
    y_edge_loc_valid = y_edge_loc(idx_valid);
    y_edge_loc_test = y_edge_loc(idx_test);

    x_train = cat(4, x_center_train, x_donut_train, x_edge_loc_train);
    x_valid = cat(4, x_center_valid, x_donut_valid, x_edge_loc_valid);
    x_test = cat(4, x_center_test, x_donut_test, x_edge_loc_test);
    y_train = cat(1, y_center_train, y_donut_train, y_edge_loc_train);
    y_valid = cat(1, y_center_valid, y_donut_valid, y_edge_loc_valid);
    y_test = cat(1, y_center_test, y_donut_test, y_edge_loc_test);
    
    [n_per_class, edges] = histcounts(y_train);
    classes = [0 1 2];
    max_class = max(n_per_class);
    add_per_class = max_class - n_per_class;
    idx_added = [];
    for c = 1:3
        added = 0;
        while added < add_per_class(c)
            idx_add = randi(size(y_train, 1));
            if y_train(idx_add) == categorical(classes(c))
                idx_added(end + 1) = idx_add;
                added = added + 1;
            end
        end
    end
    x_train_added = rotate(x_train(:, :, :, idx_added));
    y_train_added = y_train(idx_added);
    x_train_aug = cat(4, x_train, x_train_added);
    y_train_aug = cat(1, y_train, y_train_added);
    
    net = trainNetwork(x_train_aug, y_train_aug, layers, options);
    y_pred = classify(net, x_test);
    test_accuracy = sum(y_pred == y_test) / numel(y_test);
    acc_aug(end + 1) = test_accuracy;
end

scatter(0.05:0.05:0.8, acc)
hold on
scatter(0.05:0.05:0.8, acc_aug)
legend('Without Augmentation', 'With Augmentation')
hold off