%% Loading the data
clear
clc
load('wafer_maps.mat')

% Extracting the number of samples for each class
n_center = size(center, 4);
n_donut = size(donut, 4);
n_edge_loc = size(edge_loc, 4);

%% Creating the label vectors
y_center = categorical(zeros(n_center, 1));
y_donut = categorical(ones(n_donut, 1));
y_edge_loc = categorical(ones(n_edge_loc, 1) * 2);

%% Preparing training, validation, and test sets
% Assuming you want to take 50% of each class for training, 25% for validation, and 25% for testing

% Define the percentages
percentage_train = 50;
percentage_validation = 25;
percentage_test = 25;

% Calculate the number of samples to take for each class
num_train_center = round(percentage_train / 100 * n_center);
num_train_donut = round(percentage_train / 100 * n_donut);
num_train_edge_loc = round(percentage_train / 100 * n_edge_loc);

num_validation_test_center = round(percentage_validation / 100 * n_center);
num_validation_test_donut = round(percentage_validation / 100 * n_donut);
num_validation_test_edge_loc = round(percentage_validation / 100 * n_edge_loc);

% Shuffle and split the data for training, validation, and test
idx_center = randperm(n_center);
idx_donut = randperm(n_donut);
idx_edge_loc = randperm(n_edge_loc);

% Training set
x_train = cat(4, center(:, :, :, idx_center(1:num_train_center)), ...
                    donut(:, :, :, idx_donut(1:num_train_donut)), ...
                    edge_loc(:, :, :, idx_edge_loc(1:num_train_edge_loc)));

y_train = cat(1, y_center(1:num_train_center), ...
                 y_donut(1:num_train_donut), ...
                 y_edge_loc(1:num_train_edge_loc));

% Validation set
x_validation = cat(4, center(:, :, :, idx_center(num_train_center + 1 : num_train_center + num_validation_test_center)), ...
                         donut(:, :, :, idx_donut(num_train_donut + 1 : num_train_donut + num_validation_test_donut)), ...
                         edge_loc(:, :, :, idx_edge_loc(num_train_edge_loc + 1 : num_train_edge_loc + num_validation_test_edge_loc)));

y_validation = cat(1, y_center(num_train_center + 1 : num_train_center + num_validation_test_center), ...
                      y_donut(num_train_donut + 1 : num_train_donut + num_validation_test_donut), ...
                      y_edge_loc(num_train_edge_loc + 1 : num_train_edge_loc + num_validation_test_edge_loc));

% Test set
x_test = cat(4, center(:, :, :, idx_center(num_train_center + num_validation_test_center + 1 : end)), ...
                   donut(:, :, :, idx_donut(num_train_donut + num_validation_test_donut + 1 : end)), ...
                   edge_loc(:, :, :, idx_edge_loc(num_train_edge_loc + num_validation_test_edge_loc + 1 : end)));

y_test = cat(1, y_center(num_train_center + num_validation_test_center + 1 : end), ...
                    y_donut(num_train_donut + num_validation_test_donut + 1 : end), ...
                    y_edge_loc(num_train_edge_loc + num_validation_test_edge_loc + 1 : end));

%% Model definition and training
% Your CNN model definition and training code here
rng(3)
layers = [
    imageInputLayer([32 32 1])
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,64,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
    	
    fullyConnectedLayer(3)
    softmaxLayer
    classificationLayer];

miniBatchSize = 64;
options = trainingOptions( 'adam', 'MiniBatchSize', miniBatchSize, ...
    'MaxEpochs', 4, 'Shuffle', 'every-epoch', ...
    'ValidationData', {x_validation, y_validation}, 'ValidationFrequency', 10, ...
    'Plots', 'training-progress');

net = trainNetwork(x_train, y_train, layers, options);

%% Testing the model
% Your model testing code here

% Use the trained model to classify the test set
y_pred = classify(net, x_test);

% Calculate accuracy
test_accuracy = sum(y_pred == y_test) / numel(y_test);
disp(['Test Accuracy: ' num2str(test_accuracy)])

% Create a confusion matrix
confusion_mat = confusionmat(y_test, y_pred);

% Display confusion matrix
confusionchart(confusion_mat, 'RowSummary','row-normalized', 'ColumnSummary', 'column-normalized');


%% Accuracy vs. percentage data for training
% Loop through different percentages of data per class for training
% Train the model on each percentage and record the accuracy
% Plot the results

%% Accuracy vs. percentage data for training



%% Accuracy vs. percentage data for training with data augmentation
% Similar to the previous section, but perform data augmentation on the training set
% Train the model on each percentage and record the accuracy
% Plot the results
