%% Loading the data
clear
clc
close all

load('wafer_maps.mat')

imshow(center(:, :, :, 2))

n_center = size(center, 4); n_center
n_donut = size(donut, 4); n_donut
n_edge_loc = size(edge_loc, 4); n_edge_loc

%% Creating the label vectors
y_center = categorical(zeros(size(center, 4), 1));
y_donut = categorical(ones(size(donut, 4), 1));
y_edge_loc = categorical(ones(size(edge_loc, 4), 1) * 2);

%% Preparing training, validation, and test sets
% Assuming you have 3 classes represented by 4D matrices
class1_data = center; % Your 4D matrix for class 1
class2_data = donut; % Your 4D matrix for class 2
class3_data = edge_loc; % Your 4D matrix for class 3

%%
%training,test,validation dimensions
c_t=0.05:0.05:0.85;

for i=1:length(c_t)
c_v=(1-c_t(i))/2;
c_x=c_v;
%% class 1
% Get the size of the data
[height1, width1, depth1, num_samples1] = size(class1_data);

% Calculate the indices for training, validation, and test sets
train_indices1 = 1:round(c_t(i) * num_samples1);
val_indices1 = (round(c_t(i) * num_samples1) + 1):(round((c_t(i)+c_v) * num_samples1));
test_indices1 = (round((c_t(i)+c_v) * num_samples1) + 1):num_samples1;

% Extract the data for each set
train_data1 = class1_data(:, :, :, train_indices1);
val_data1 = class1_data(:, :, :, val_indices1);
test_data1 = class1_data(:, :, :, test_indices1);

y_train1=y_center(train_indices1);
y_val1=y_center(val_indices1);
y_test1=y_center(test_indices1);


%% class 2
% Get the size of the data
[height2, width2, depth2, num_samples2] = size(class2_data);

% Calculate the indices for training, validation, and test sets
train_indices2 = 1:round(c_t(i) * num_samples2);
val_indices2 = (round(c_t(i) * num_samples2) + 1):(round((c_t(i)+c_v) * num_samples2));
test_indices2 = (round((c_t(i)+c_v) * num_samples2) + 1):num_samples2;

% Extract the data for each set
train_data2 = class2_data(:, :, :, train_indices2);
val_data2 = class2_data(:, :, :, val_indices2);
test_data2 = class2_data(:, :, :, test_indices2);

y_train2=y_donut(train_indices2);
y_val2=y_donut(val_indices2);
y_test2=y_donut(test_indices2);
%% class 3
% Get the size of the data
[height3, width3, depth3, num_samples3] = size(class3_data);

% Calculate the indices for training, validation, and test sets
train_indices3 = 1:round(c_t(i) * num_samples3);
val_indices3 = (round(c_t(i) * num_samples3) + 1):(round((c_t(i)+c_v) * num_samples3));
test_indices3 = (round((c_t(i)+c_v) * num_samples3) + 1):num_samples3;

% Extract the data for each set
train_data3 = class3_data(:, :, :, train_indices3);
val_data3 = class3_data(:, :, :, val_indices3);
test_data3 = class3_data(:, :, :, test_indices3);

y_train3=y_edge_loc(train_indices3);
y_val3=y_edge_loc(val_indices3);
y_test3=y_edge_loc(test_indices3);

%% repacking
x_train=cat(4,train_data1,train_data2,train_data3);
y_train=cat(1,y_train1,y_train2,y_train3);

x_val=cat(4,val_data1,val_data2,val_data3);
y_val=cat(1,y_val1,y_val2,y_val3);

x_test=cat(4,test_data1,test_data2,test_data3);
y_test=cat(1,y_test1,y_test2,y_test3);

%shuffling 
sh_train=randperm(size(x_train,4));
x_train=x_train(:,:,:,sh_train);
y_train=y_train(sh_train);

sh_val=randperm(size(x_val,4));
x_val=x_val(:,:,:,sh_val);
y_val=y_val(sh_val);

sh_test=randperm(size(x_test,4));
x_test=x_test(:,:,:,sh_test);
y_test=y_test(sh_test);

figure
histogram(y_train)
%% Model definition and training

rng(3) %seed of random number
layers = [
    imageInputLayer([32 32 1])
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,64,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
    	
    fullyConnectedLayer(3)
    softmaxLayer
    classificationLayer];

miniBatchSize = 64;
options = trainingOptions( 'adam', 'MiniBatchSize', miniBatchSize, ...
    'MaxEpochs', 10, 'Shuffle', 'every-epoch', ...
    'ValidationData', {x_val, y_val}, 'ValidationFrequency', 10);

net = trainNetwork(x_train, y_train, layers, options);

%% Testing the model
y_pred = classify(net, x_test);
test_accuracy(i) = sum(y_pred == y_test) / numel(y_test);
test_accuracy(i)
figure
confusionchart(y_test, y_pred)




%% Accuracy vs. percentage data for training with data augmentation
% Data augmentation
% The classes are not equal in bin size, so we random increase the number
% of images in an artifical way: take a high quality image of a given digit
% and add artificial noise and rotation.
rng(3)
[n_per_class, edges] = histcounts(y_train);
classes = [0 1 2];
max_class = max(n_per_class);
add_per_class = max_class - n_per_class;

idx_added = [];
for c = 1:3
    added = 0;
    while added < add_per_class(c)
        idx_add = randi(size(y_train, 1));
        if y_train(idx_add) == categorical(classes(c))
            idx_added(end + 1) = idx_add;
            added = added + 1;
        end
    end
end

x_train_added = add_noise(x_train(:, :, :, idx_added));
y_train_added = y_train(idx_added);
x_train_aug = cat(4, x_train, x_train_added);
y_train_aug = cat(1, y_train, y_train_added);
figure
histogram(y_train_aug)

%% Adding shifting to Augmentation

rng(3)
x_train_added = add_noise(shift(x_train(:, :, :, idx_added)));
x_train_aug = cat(4, x_train, x_train_added);

net = trainNetwork(x_train_aug, y_train_aug, layers, options);

y_pred = classify(net, x_test);
test_accuracy_aug(i) = sum(y_pred == y_test) / numel(y_test);

figure
confusionchart(y_test, y_pred)
end
%% Accuracy vs. percentage data for training
figure
scatter(c_t,test_accuracy)
hold on;
scatter(c_t,test_accuracy_aug)
legend('Starting Dataset','Augmented and Shifted Dataset')